"""
Sends a few testing messages to the configured syslog server.
"""
import argparse
import socket

def get_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(description="Sends 3 testing messages to the configured syslog server.")
    parser.add_argument('--host', metavar="HOST",help="Syslog server host.", required=True)
    parser.add_argument('--port', metavar="PORT", help="Syslog server port.", required=True)
    parser.add_argument('--protocol', metavar="'TCP' or 'UDP'", help="Syslog server protocol.", required=True)
    return parser

if __name__ == "__main__":
    from syslogcef import SyslogCEFSender, __version__

    args = get_parser().parse_args()

    syslog = SyslogCEFSender(
        host=args.host, 
        port=args.port, 
        protocol=args.protocol, 
        deviceProduct='syslogcef.testmessages', 
        deviceVendor='syslogcef',
        deviceVersion=__version__.__version__, 
        sourceHostName=socket.gethostname(), )

    syslog.register_event('1113-100', name='Testing syslogcef OK', severity=0)
    syslog.register_event('1113-101', name='Testing syslogcef WARN', severity=5)
    syslog.register_event('1113-102', name='Testing syslogcef ERROR', severity=9, 
        reason="This is only a test.")

    syslog.send('1113-100', message="The syslogcef module is working properly.")
    syslog.send('1113-101', message="The syslogcef module is still working properly.")
    syslog.send('1113-102', message="The syslogcef module is working properly again.")



__title__ = "syslogcef"

__description__ = "Python library to easily send CEF formatted messages to syslog server"

__url__ = "https://github.com/tristanlatr/syslogcef"

__license__ = "MIT License"

__keywords__ = "syslog cef"

__version__ = "0.1.8"



import typing as t
import pytest

from syslogcef import CEFSender, StdoutSyslogSender, CompositeSyslogSender

if t.TYPE_CHECKING:
    class CapLog(t.Protocol):
        text:str
    
    class CaptureResult(t.Protocol):
        out: str
        err: str

    class CapSys(t.Protocol):
        def readouterr(self) -> CaptureResult: 
            ...
else:
    CapLog = object
    CapSys = object

def test_syslogcef(capsys:CapSys) -> None:

    s = CEFSender(StdoutSyslogSender(), 
        deviceProduct='test-syslogcef', 
        deviceVendor='Github',
        deviceVersion='14.1')

    s.register_event('100', 'Device is reachable', 1)
    s.register_event('101', 'Device is slow to respond', 5)
    s.register_event('102', 'Device is down', 10, reason="Ping failed")

    s.send('100', message='Ping time is: 33ms')
    s.send('101', message='Ping time is: 521ms')
    s.send('102', message='Unreachable')

    assert capsys.readouterr().out == """CEF:0|Github|test-syslogcef|14.1|100|Device is reachable|1|msg=Ping time is: 33ms
CEF:0|Github|test-syslogcef|14.1|101|Device is slow to respond|5|msg=Ping time is: 521ms
CEF:0|Github|test-syslogcef|14.1|102|Device is down|10|reason=Ping failed msg=Unreachable
"""

def test_syslogcef_defaults(capsys:CapSys) -> None:

    s = CEFSender(StdoutSyslogSender(),)

    s.register_event('100', 'Device is reachable', 1)
    s.register_event('101', 'Device is slow to respond', 5)
    s.register_event('102', 'Device is down', 10, reason="Ping failed")

    s.send('100', message='Ping time is: 33ms')
    s.send('101', message='Ping time is: 521ms')
    s.send('102', message='Unreachable')

    assert capsys.readouterr().out == """CEF:0|CEF Vendor|CEF Product|1.0|100|Device is reachable|1|msg=Ping time is: 33ms
CEF:0|CEF Vendor|CEF Product|1.0|101|Device is slow to respond|5|msg=Ping time is: 521ms
CEF:0|CEF Vendor|CEF Product|1.0|102|Device is down|10|reason=Ping failed msg=Unreachable
"""

def test_syslogcef_composite_logger(capsys:CapSys) -> None:
    s = CEFSender(CompositeSyslogSender(StdoutSyslogSender(),StdoutSyslogSender()))
    s.register_event('100', 'Device is reachable', 1)
    s.send('100', message='Ping time is: 33ms')
    assert capsys.readouterr().out == """CEF:0|CEF Vendor|CEF Product|1.0|100|Device is reachable|1|msg=Ping time is: 33ms
CEF:0|CEF Vendor|CEF Product|1.0|100|Device is reachable|1|msg=Ping time is: 33ms
"""

def test_invalid_cef_field() -> None:
    s = CEFSender(StdoutSyslogSender(),)
    with pytest.raises(ValueError):
        s.register_event('100', 'Device is reachable', 1, host='name')
    with pytest.raises(ValueError):
        s.register_event('100', 'Device is reachable', 1, port='port')
    with pytest.raises(ValueError):
        s.register_event('100', 'Device is reachable', 1, port=443)
    with pytest.raises(ValueError):
        s.register_event('100', 'Device is reachable', 1, protocol=443)
    with pytest.raises(ValueError):
        s.register_event('100', 'Device is reachable', 1, protocol='https')


#! /usr/bin/env python3
from setuptools import setup, find_packages
import pathlib
# The directory containing this file
HERE = pathlib.Path(__file__).parent
# About the project
ABOUT = {}
exec((HERE / "syslogcef" / "__version__.py").read_text(), ABOUT)
# The text of the README file
README = (HERE / "README.md").read_text()
setup(
    name                =   ABOUT['__title__'],
    description         =   ABOUT['__description__'],
    url                 =   ABOUT['__url__'],
    version             =   ABOUT['__version__'],
    packages            =   find_packages(exclude=('tests')), 
    classifiers         =   ["Programming Language :: Python :: 3",],
    license             =   ABOUT['__license__'],
    long_description    =   README,
    long_description_content_type   =   "text/markdown",
    python_requires     =   '>=3.6',
    install_requires    =   ['rfc5424-logging-handler', 
                             'cefevent>=0.5.3',],
    extras_require      =   {'dev': ["tox"]},
    keywords            =   ABOUT['__keywords__'],
)



import logging
import socket
import uuid
from typing import Any, Dict, TYPE_CHECKING, Type

if TYPE_CHECKING:
    from typing import Protocol
else:
    Protocol = object

from rfc5424logging import Rfc5424SysLogHandler
from cefevent import CEFEvent
from cefevent.syslog import Syslog

class SyslogSenderFactory(Protocol):
    def __call__(self, host:str, port:int=514, protocol:str="UDP") -> 'SyslogSender':
        ...

class SyslogSender(Protocol):
    def send(self, msg:str) -> None:...

class CEFSender:
    """
    Base class to send CEF messages.
    """

    class EventMeta:
        """
        Wrap a CEF event definition.
        """

        def __init__(self, signatureId:str, 
                    name:str, 
                    severity:int, 
                    **fields:Any) -> None:

            self.fields = dict(
                signatureId = signatureId,
                name = name,
                severity = severity,
            )
            if fields:
                self.fields.update(fields)

    def __init__(self, syslog_sender:SyslogSender, **fields:Any ) -> None:
        """
        Create a CEFSender instance. 
        
        You might want to use SyslogCEFSender to create a sender from syslog server hostname directly.
        """
        
        self.syslog_sender = syslog_sender

        self.fields = {}
        if fields:
            self.fields.update(fields)

        self.registered_events:Dict[str, CEFSender.EventMeta] = {}

    def register_event(self, signatureId:str, name:str, severity:int, **fields:Any) -> None:
        """
        Register a new event definition.
        """
        self.registered_events[signatureId] = CEFSender.EventMeta(
            signatureId=signatureId, 
            name=name, 
            severity=severity,
            **fields
        )
        
        # try to build a CEF message to raise ValueError if the Event definition is invalid
        self._build_cef(signatureId)

    def _build_cef(self, signatureId:str, **fields:Any) -> str:
        event_meta = self.registered_events[signatureId]

        _fields: Dict[str, Any] = {}
        _fields.update(self.fields)
        _fields.update(event_meta.fields)
        _fields.update(fields)

        cef = CEFEvent(strict=True)

        for k,v in _fields.items():
            try:
                cef.set_field(k,v)
            except ValueError:
                if isinstance(v, str):
                    cef.set_field(k,v[:1022])
                else:
                    raise

        msg = cef.build_cef()
        assert isinstance(msg, str)
        return msg

    def send(self, signatureId:str, **fields:Any) -> None:
        """
        Send a CEF message. 

        Args:
            signatureId: ID of the registered event.
            **fields: Additional fields to include into the CEF message, like ::
                message="Error #28", sourceHostName="127.0.0.1"
        
        Raises:
            ValueError: If the a field has invalid value.
        """
        self.syslog_sender.send(self._build_cef(signatureId, **fields))

class Rfc3164SyslogSender(SyslogSender):
    """
    `rfc3164 <https://datatracker.ietf.org/doc/html/rfc3164>`_ sender.
    """
    def __init__(self, host: str, port: int = 514, protocol: str = "UDP"):
        assert protocol in ["TCP", "UDP"], f"Invalid protocol {protocol!r}, please choose 'TCP' or 'UDP'."
        self.host = host
        self.port = port
        self.protocol = protocol
        
        self.syslog = Syslog(host=host, port=port, protocol=protocol)
    
    def send(self, msg: str) -> None:
        self.syslog.notice(msg)

cefeventSyslogSender = Rfc3164SyslogSender # for compatibility, do not use in new code

class Rfc5424SyslogSender(SyslogSender):
    """
    `rfc5424 <https://datatracker.ietf.org/doc/html/rfc5424>`_ sender.
    """
    def __init__(self, host: str, port: int = 514, protocol: str = "UDP", **kwargs:Any):
        """
        :param kwargs: Any extra arguments are forwarded to Rfc5424SysLogHandler instance.
        """
        assert protocol in ["TCP", "UDP"], f"Invalid protocol {protocol!r}, please choose 'TCP' or 'UDP'."

        logger_name = f'syslogcef-{uuid.uuid4()}'
        self.logger = logging.getLogger(logger_name)

        init_kwargs = dict(
            appname='syslogcef',
            enterprise_id=42, 
            msg_as_utf8=True, 
            utc_timestamp=True
        )
        init_kwargs.update(**kwargs)
        
        sh: Rfc5424SysLogHandler = Rfc5424SysLogHandler(
            address=(host, port),
            socktype=socket.SOCK_STREAM if protocol=='TCP' else socket.SOCK_DGRAM,  # Use TCP or UDP
            **init_kwargs,
        )
        
        self.logger.setLevel(logging.DEBUG)
        self.logger.addHandler(sh)
    
    def send(self, msg: str) -> None:
        self.logger.info(msg)

class StdoutSyslogSender(SyslogSender):
    """
    Send CEF messages to stdout.
    """
    def send(self, msg:str) -> None:
        print(msg)

class CompositeSyslogSender(SyslogSender):
    """
    Send CEF messages to several underlying senders.
    """
    def __init__(self, *senders:SyslogSender) -> None:
        self.senders = senders
    def send(self, msg:str) -> None:
        for sender in self.senders:
            sender.send(msg)

class SyslogCEFSender(CEFSender):
    """
    Main object to easily send CEF messages to a syslog server.

    You might want to use CEFSender to have more flexibility.
    """

    def __init__(self, host: str, 
                port: int,
                protocol:str,
                syslog_sender_class:SyslogSenderFactory=Rfc5424SyslogSender,
                **fields: Any) -> None:
        """
        Create a SyslogCEFSender.
        """
        super().__init__(syslog_sender_class(host=host, port=port, protocol=protocol), **fields)
